def test_rust_alias_imports():
    from rust_circuit import (
        Axis,
        Binder,
        CircuitColorer,
        CircuitHtmlColorer,
        CliColor,
        GeneralFunctionSpec,
        IntOrMatcher,
        IrreducibleNode,
        IterativeMatcherIn,
        Leaf,
        LeafConstant,
        MatcherIn,
        MaybeUpdate,
        ModuleConstructCallback,
        NestAddsSpec,
        NestAddsSpecMultiple,
        NestAddsSpecSub,
        NestedModuleNamer,
        NestEinsumsSpec,
        NestEinsumsSpecMultiple,
        NestEinsumsSpecSub,
        PrintOptionsBase,
        SampleSpecIn,
        Shape,
        TorchAxisIndex,
        TransformIn,
        UpdatedIterativeMatcher,
        UpdatedIterativeMatcherIn,
        Var,
    )

    _ = (
        Shape,
        Axis,
        IrreducibleNode,
        Leaf,
        LeafConstant,
        Var,
        MatcherIn,
        IterativeMatcherIn,
        TransformIn,
        SampleSpecIn,
        TorchAxisIndex,
        IntOrMatcher,
        NestEinsumsSpecMultiple,
        NestEinsumsSpecSub,
        NestEinsumsSpec,
        NestAddsSpecMultiple,
        NestAddsSpecSub,
        NestAddsSpec,
        GeneralFunctionSpec,
        Binder,
        UpdatedIterativeMatcher,
        UpdatedIterativeMatcherIn,
        PrintOptionsBase,
        CliColor,
        CircuitColorer,
        CircuitHtmlColorer,
        NestedModuleNamer,
        ModuleConstructCallback,
        MaybeUpdate,
    )
