#![feature(let_chains)]
#![feature(iterator_try_reduce)]
pub mod cumulant_rewrites;
pub mod diff;
pub mod distribute_and_factor;
pub mod generalfunction;
pub mod index_rewrites;
pub mod modules;
pub mod nest;
